//
//  FirmwareUpdateViewController.m
//  Bluzone
//
//  Created by Jason Clary on 7/28/16.
//  Copyright © 2016 Bluvision. All rights reserved.
//

#import "FirmwareUpdater.h"
#import "FirmwareStorage.h"
#import "MRProgress.h"
#import "MRProgressOverlayView.h"
#import <Bluvision/Bluvision.h>


@interface FirmwareUpdater () <BLUConfigurableBeaconDelegate>

@property (nonatomic, strong) FirmwareStorage *firmwareStorage;
@property (nonatomic, strong) MRProgressOverlayView *progressView;

@property (weak, nonatomic) id<BLUConfigurableBeaconDelegate> previousDelegate;

@end

@implementation FirmwareUpdater

- (instancetype)init {
    self = [super init];

    if (!self) {
        return nil;
    }

    self.firmwareStorage = [FirmwareStorage defaultStorage];

    
    return self;
}


- (void)beginFirmwareUpdate {
    self.progressView = [MRProgressOverlayView showOverlayAddedTo:[UIApplication sharedApplication].keyWindow animated:YES];
    self.progressView.mode = MRProgressOverlayViewModeDeterminateCircular;
    self.progressView.titleLabelText = @"Updating";

    [self updateFirmware];
}

- (void)updateFirmware {
    if (self.configurableBeacon.connectionState != BLUConfigurableBeaconConnectionStateConnected) {
        self.progressView.mode = MRProgressOverlayViewModeIndeterminate;
        self.progressView.titleLabelText = @"Connecting";
        [self reconnectAndContiuneFirmwareUpdate];
        return;
    }

    if (![self.configurableBeacon isKindOfClass:[BLUSBeacon class]] &&
        ![self.configurableBeacon isKindOfClass:[BLUSLegacyBeacon class]] &&
        ![self.configurableBeacon isKindOfClass:[BLUSEncryptedBeacon class]]) {

        [MRProgressOverlayView dismissOverlayForView:[UIApplication sharedApplication].keyWindow animated:YES];
        return;
    }

    BLUSBeacon *sBeacon = (BLUSBeacon *)self.configurableBeacon;
    BLUBeaconFirmwareUpdateStage stage = [self.firmwareStorage storedStageForConfigurableBeacon:self.configurableBeacon];

    BLUFirmwareImage *image = self.stage1FirmwareImage;
    if (stage == BLUBeaconFirmwareUpdateStageNone) {
        // Fresh update ...
        if (image) {
            if (image.isLoaderImage) {
                stage = BLUBeaconFirmwareUpdateStageWipe;
            }
            else {
                image = self.stage2FirmwareImage ?: [self.firmwareStorage storedFirmwareImageForConfigurableBeacon:self.configurableBeacon];
                stage = BLUBeaconFirmwareUpdateStageUpdate;
            }
        }
        else {
            stage = BLUBeaconFirmwareUpdateStageUpdate;
            image = self.stage2FirmwareImage ?: [self.firmwareStorage storedFirmwareImageForConfigurableBeacon:self.configurableBeacon];
            if (![self checkFirmwareImageAndFail:image stage:stage]) {
                return;
            }

            if ([sBeacon requiresLoaderImageForFirmwareImage:image]) {
                stage = BLUBeaconFirmwareUpdateStageWipe;
                image = self.stage1FirmwareImage;
                if (![self checkFirmwareImageAndFail:image stage:stage]) {
                    return;
                }
            }
        }
        [self startUpdatingFirmwareWithImage:image stage:stage isFirstUpdate:YES];
    }
    else if (stage == BLUBeaconFirmwareUpdateStageWipe) {
//        if ([self.configurableBeacon.name hasSuffix:@"Loader"]) { // This is not a real fix. 
//            image = self.stage2FirmwareImage ?: [self.firmwareStorage storedFirmwareImageForConfigurableBeacon:self.configurableBeacon];
//            [self startUpdatingFirmwareWithImage:image stage:BLUBeaconFirmwareUpdateStageUpdate isFirstUpdate:NO];
//            return;
//        }

        // Wiping failed before? Repeat ...
        if (!image || !image.isLoaderImage) {
            image = self.stage1FirmwareImage;
        }
        if (![self checkFirmwareImageAndFail:image stage:stage]) {
            return;
        }
        [self startUpdatingFirmwareWithImage:image stage:stage isFirstUpdate:NO];
    }
    else if (stage == BLUBeaconFirmwareUpdateStageUpdate) {
        // Wiping successful, continue updating ...
        //if (!image || image.isLoaderImage) {
        //    image = [self stageTwoFirmwareImageForSBeacon:sBeacon];
        //}
        image = self.stage2FirmwareImage ?: [self.firmwareStorage storedFirmwareImageForConfigurableBeacon:self.configurableBeacon];


        if (![self checkFirmwareImageAndFail:image stage:stage]) {
            return;
        }
        [self startUpdatingFirmwareWithImage:image stage:stage isFirstUpdate:NO];
    }
}

- (void)startUpdatingFirmwareWithImage:(BLUFirmwareImage *)image stage:(BLUBeaconFirmwareUpdateStage)stage isFirstUpdate:(BOOL)firstUpdate {

    [self.firmwareStorage storeFirmwareUpdateStage:stage image:self.stage2FirmwareImage forConfigurableBeacon:self.configurableBeacon];
    if (firstUpdate) {
        [self.firmwareStorage storeConfiguration:self.configurableBeacon.configuration forConfigurableBeacon:self.configurableBeacon];
    }
    [self.configurableBeacon updateFirmwareWithImage:image progress:^(float percentComplete) {
        if (self.progressView.mode != MRProgressOverlayViewModeDeterminateCircular) {
            self.progressView.mode = MRProgressOverlayViewModeDeterminateCircular;
        }

        [self.progressView setProgress:((float)percentComplete / 100.0) animated:YES];
    } completion:^(BOOL success, NSError *error) {
        NSLog(@"Firmware Update completed with error: %@", error.localizedDescription);

        [self notifyFirmwareUpdateFinishedWithError:error stage:stage];
    }];
}

- (void)notifyFirmwareUpdateFinishedWithError:(NSError *)error stage:(BLUBeaconFirmwareUpdateStage)stage {

    self.progressView.mode = MRProgressOverlayViewModeCheckmark;
    if (error) {
        self.progressView.titleLabelText = [NSString stringWithFormat:@"Error: %@", error.localizedDescription];

        self.configurableBeacon.delegate = self.previousDelegate;

        if ([self.delegate respondsToSelector:@selector(firmwareUpdater:didFinishFirmwareUpdateWithError:)]) {
            [self.delegate firmwareUpdater:self didFinishFirmwareUpdateWithError:error];
        }

        [MRProgressOverlayView dismissOverlayForView:[UIApplication sharedApplication].keyWindow animated:YES];
    }
    else if (stage == BLUBeaconFirmwareUpdateStageWipe) {
        stage = BLUBeaconFirmwareUpdateStageUpdate;
        [self.firmwareStorage storeFirmwareUpdateStage:stage image:self.stage2FirmwareImage forConfigurableBeacon:self.configurableBeacon];
        self.progressView.titleLabelText = @"Stage 1 Complete: Reconnecting";

        dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(2 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^{
            [self reconnectAndContiuneFirmwareUpdate];
        });
    }
    else {
        self.progressView.titleLabelText = @"Firmware update completed";
        [self.firmwareStorage removeStoredStageForConfigurableBeacon:self.configurableBeacon];

        if ([self.delegate respondsToSelector:@selector(firmwareUpdater:didFinishFirmwareUpdateWithError:)]) {
            [self.delegate firmwareUpdater:self didFinishFirmwareUpdateWithError:nil];
        }

        self.configurableBeacon.delegate = self.previousDelegate;

        [MRProgressOverlayView dismissOverlayForView:[UIApplication sharedApplication].keyWindow animated:YES];
    }
}

- (void)reconnectAndContiuneFirmwareUpdate {
    self.previousDelegate = self.configurableBeacon.delegate;
    self.configurableBeacon.delegate = self;
    [self.configurableBeacon connect];
}

- (void)configurableBeaconDidConnect:(BLUConfigurableBeacon *)configurableBeacon {
    self.progressView.titleLabelText = @"Updating";
    [self updateFirmware];
}

- (void)configurableBeacon:(BLUConfigurableBeacon *)configurableBeacon didDisconnectWithError:(NSError *)error {
    if (error) {
        NSLog(@"%s - %@", __PRETTY_FUNCTION__, error);
        [self showAlertWithMessage:error.localizedDescription];
        [MRProgressOverlayView dismissOverlayForView:[UIApplication sharedApplication].keyWindow animated:YES];
    }
}

- (BOOL)checkFirmwareImageAndFail:(BLUFirmwareImage *)image stage:(BLUBeaconFirmwareUpdateStage)stage {
    if (!image) {
        NSError *error = [[NSError alloc] initWithDomain:@"BLUBeaconTestAppErrorDomain" code:1 userInfo:@{NSLocalizedDescriptionKey : @"No image found for firmware update"}];
        [self notifyFirmwareUpdateFinishedWithError:error stage:stage];

        return NO;
    }
    return YES;
}

- (void)showAlertWithMessage:(NSString *)message {
    UIAlertController *alert = [UIAlertController alertControllerWithTitle:@"Mesage" message:message preferredStyle:UIAlertControllerStyleAlert];

    [alert addAction:[UIAlertAction actionWithTitle:@"OK" style:UIAlertActionStyleDefault handler:nil]];

    [[[UIApplication sharedApplication] delegate].window.rootViewController presentViewController:alert animated:YES completion:nil];
}


@end
